@@ -7,7 +7,7 @@
 
 """Parse DDL statements"""
 
-import re
+import re, textwrap, json
 from collections import OrderedDict
 from enum import IntEnum
 
@@ -27,6 +27,7 @@ class DdlParseBase():
     def source_database(self):
         """
         Source database option
+
         :param source_database: enum DdlParse.DATABASE
         """
         return self._source_database
@@ -51,14 +52,14 @@ class DdlParseTableColumnBase(DdlParseBase):
     def name(self, name):
         self._name = name
 
-    def _get_name(self, name_case=DdlParseBase.NAME_CASE.original):
+    def get_name(self, name_case=DdlParseBase.NAME_CASE.original):
         """
         Get Name converted case
 
         :param name_case: name case type
-            * NAME_CASE.original : Return to no convert
-            * NAME_CASE.lower : Return to lower
-            * NAME_CASE.upper : Return to upper
+            * DdlParse.NAME_CASE.original : Return to no convert
+            * DdlParse.NAME_CASE.lower : Return to lower
+            * DdlParse.NAME_CASE.upper : Return to upper
 
         :return: name
         """
@@ -73,16 +74,19 @@ class DdlParseTableColumnBase(DdlParseBase):
 class DdlParseColumn(DdlParseTableColumnBase):
     """Column define info"""
 
-    def __init__(self, name, data_type_array, constraint=None, source_database=None):
+    def __init__(self, name, data_type_array, array_brackets=None, constraint=None, source_database=None):
         """
         :param data_type_array[]: Column data type ['data type name'] or ['data type name', '(length)'] or ['data type name', '(precision, scale)']
+        :param array_brackets: Column array brackets string '[]' or '[][]...'
         :param constraint: Column constraint string
         :param source_database: enum DdlParse.DATABASE
         """
+
         super().__init__(source_database)
         self._name = name
         self._set_data_type(data_type_array)
         self.constraint = constraint
+        self._array_dimensional = 0 if array_brackets is None else array_brackets.count('[]')
 
     @property
     def data_type(self):
@@ -108,11 +112,12 @@ class DdlParseColumn(DdlParseTableColumnBase):
         if len(data_type_array) < 2:
             return
 
-        matches = re.findall(r"(\d+)\s*,*\s*(\d*)", data_type_array[1])
+        matches = re.findall(r"([\d\*]+)\s*,*\s*(\d*)", data_type_array[-1])
         if len(matches) > 0:
-            self._length = int(matches[0][0])
+            self._length = matches[0][0] if matches[0][0] == "*" else int(matches[0][0])
             self._scale = None if len(matches[0]) < 2 or matches[0][1] == "" or int(matches[0][1]) == 0 else int(matches[0][1])
-        else:
+
+        if re.search(r"^[^\d\*]+", data_type_array[1]):
             self._data_type += " {}".format(data_type_array[1])
 
 
@@ -135,6 +140,33 @@ class DdlParseColumn(DdlParseTableColumnBase):
         self._pk = False if self._constraint is None or not re.search("PRIMARY KEY", self._constraint) else True
         self._unique = False if self._constraint is None or not re.search("UNIQUE", self._constraint) else True
 
+        self._comment = None
+        if constraint is not None:
+            matches = re.findall(r"(?:\bCOMMENT\b\s+)(['\"])(.+)\1", constraint, re.IGNORECASE)
+            if len(matches) > 0:
+                self._comment = matches[0][1]
+
+    @property
+    def comment(self):
+        return self._comment
+
+    @comment.setter
+    def comment(self, comment):
+        self._comment = comment
+
+    @property
+    def description(self):
+        return self.comment
+
+    @description.setter
+    def description(self, description):
+        self.comment = description
+
+    @property
+    def array_dimensional(self):
+        """array dimensional number"""
+        return self._array_dimensional
+
     @property
     def not_null(self):
         return self._not_null
@@ -161,11 +193,11 @@ class DdlParseColumn(DdlParseTableColumnBase):
 
     @property
     def bigquery_data_type(self):
-        """Get BigQuery data type"""
+        """Get BigQuery Legacy SQL data type"""
 
         # BigQuery data type = {source_database: [data type, ...], ...}
         BQ_DATA_TYPE_DIC = OrderedDict()
-        BQ_DATA_TYPE_DIC["STRING"] = {None: [re.compile(r"(CHAR|TEXT|CLOB)")]}
+        BQ_DATA_TYPE_DIC["STRING"] = {None: [re.compile(r"(CHAR|TEXT|CLOB|JSON|UUID)")]}
         BQ_DATA_TYPE_DIC["INTEGER"] = {None: [re.compile(r"INT|SERIAL|YEAR")]}
         BQ_DATA_TYPE_DIC["FLOAT"] = {None: [re.compile(r"(FLOAT|DOUBLE)"), "REAL", "MONEY"]}
         BQ_DATA_TYPE_DIC["DATETIME"] = {
@@ -193,28 +225,94 @@ class DdlParseColumn(DdlParseTableColumnBase):
                         return bq_type
 
         if self._data_type in ["NUMERIC", "NUMBER", "DECIMAL"]:
-            if self._scale is not None:
-                return "FLOAT"
+            if self._length is None:
+                if self._source_database in [self.DATABASE.oracle, self.DATABASE.postgresql]:
+                    return "NUMERIC"
+                else:
+                    return "INTEGER"
 
-            if self._data_type == "NUMBER" \
-                and self._source_database == self.DATABASE.oracle \
-                and self._length is None:
-                return "FLOAT"
+            if self._length == "*":
+                return "NUMERIC"
 
-            return "INTEGER"
+            if self._length < 19:
+                if self._scale is None or self._scale == 0:
+                    return "INTEGER"
+                else:
+                    return "FLOAT"
+
+            return "NUMERIC"
 
         raise ValueError("Unknown data type : '{}'".format(self._data_type))
 
+    @property
+    def bigquery_legacy_data_type(self):
+        """Get BigQuery Legacy SQL data type"""
+
+        return self.bigquery_data_type
+
+    @property
+    def bigquery_standard_data_type(self):
+        """Get BigQuery Standard SQL data type"""
+
+        legacy_data_type = self.bigquery_data_type
+
+        if legacy_data_type == "INTEGER":
+            return "INT64"
+        elif legacy_data_type == "FLOAT":
+            return "FLOAT64"
+        elif legacy_data_type == "BOOLEAN":
+            return "BOOL"
+
+        return legacy_data_type
+
     @property
     def bigquery_mode(self):
         """Get BigQuery constraint"""
 
-        return "REQUIRED" if self.not_null else "NULLABLE"
+        if self.array_dimensional > 0:
+            return "REPEATED"
+        elif self.not_null:
+            return "REQUIRED"
+        else:
+            return "NULLABLE"
 
     def to_bigquery_field(self, name_case=DdlParseBase.NAME_CASE.original):
         """Generate BigQuery JSON field define"""
 
-        return '{{"name": "{}", "type": "{}", "mode": "{}"}}'.format(self._get_name(name_case), self.bigquery_data_type, self.bigquery_mode)
+        col_name = self.get_name(name_case)
+        mode = self.bigquery_mode
+
+        if self.array_dimensional <= 1:
+            # no or one dimensional array data type
+            type = self.bigquery_legacy_data_type
+
+        else:
+            # multiple dimensional array data type
+            type = "RECORD"
+
+            fields = OrderedDict()
+            fields_cur = fields
+
+            for i in range(1, self.array_dimensional):
+                is_last = True if i == self.array_dimensional - 1 else False
+
+                fields_cur['fields'] = [OrderedDict()]
+                fields_cur = fields_cur['fields'][0]
+
+                fields_cur['name'] = "dimension_{}".format(i)
+                fields_cur['type'] = self.bigquery_legacy_data_type if is_last else "RECORD"
+                fields_cur['mode'] = self.bigquery_mode if is_last else "REPEATED"
+
+        col = OrderedDict()
+        col['name'] = col_name
+        col['type'] = type
+        col['mode'] = mode
+        if self.description is not None:
+            col['description'] = self.description
+        if self.array_dimensional > 1:
+            col['fields'] = fields['fields']
+
+        return json.dumps(col, ensure_ascii=False)
 
 
 class DdlParseColumnDict(OrderedDict, DdlParseBase):
@@ -236,16 +334,25 @@ class DdlParseColumnDict(OrderedDict, DdlParseBase):
     def __setitem__(self, key, value):
         super().__setitem__(key.lower(), value)
 
-    def append(self, column_name, data_type_array=None, constraint=None, source_database=None):
+    def append(self, column_name, data_type_array=None, array_brackets=None, constraint=None, source_database=None):
         if source_database is None:
             source_database = self.source_database
 
-        column = DdlParseColumn(column_name, data_type_array, constraint, source_database)
+        column = DdlParseColumn(column_name, data_type_array, array_brackets, constraint, source_database)
         self.__setitem__(column_name, column)
         return column
 
     def to_bigquery_fields(self, name_case=DdlParseBase.NAME_CASE.original):
-        """Generate BigQuery JSON fields define"""
+        """
+        Generate BigQuery JSON fields define
+
+        :param name_case: name case type
+            * DdlParse.NAME_CASE.original : Return to no convert
+            * DdlParse.NAME_CASE.lower : Return to lower
+            * DdlParse.NAME_CASE.upper : Return to upper
+
+        :return: BigQuery JSON fields define
+        """
 
         bq_fields = []
 
@@ -267,6 +374,7 @@ class DdlParseTable(DdlParseTableColumnBase):
     def source_database(self):
         """
         Source database option
+
         :param source_database: enum DdlParse.DATABASE
         """
         return super().source_database
@@ -300,46 +408,170 @@ class DdlParseTable(DdlParseTableColumnBase):
         return self._columns
 
     def to_bigquery_fields(self, name_case=DdlParseBase.NAME_CASE.original):
-        """Generate BigQuery JSON fields define"""
+        """
+        Generate BigQuery JSON fields define
+
+        :param name_case: name case type
+            * DdlParse.NAME_CASE.original : Return to no convert
+            * DdlParse.NAME_CASE.lower : Return to lower
+            * DdlParse.NAME_CASE.upper : Return to upper
+
+        :return: BigQuery JSON fields define
+        """
 
         return self._columns.to_bigquery_fields(name_case)
 
+    def to_bigquery_ddl(self, name_case=DdlParseBase.NAME_CASE.original):
+        """
+        Generate BigQuery CREATE TABLE statements
+
+        :param name_case: name case type
+            * DdlParse.NAME_CASE.original : Return to no convert
+            * DdlParse.NAME_CASE.lower : Return to lower
+            * DdlParse.NAME_CASE.upper : Return to upper
+
+        :return: BigQuery CREATE TABLE statements
+        """
+
+        if self.schema is None:
+            dataset = "dataset"
+        elif name_case == self.NAME_CASE.lower:
+            dataset = self.schema.lower()
+        elif name_case == self.NAME_CASE.upper:
+            dataset = self.schema.upper()
+        else:
+            dataset = self.schema
+
+        cols_defs = []
+        for col in self.columns.values():
+            col_name = col.get_name(name_case)
+
+            if col.array_dimensional < 1:
+                # no array data type
+                type = col.bigquery_standard_data_type
+                not_null = " NOT NULL" if col.not_null else ""
+
+            else:
+                # one or multiple dimensional array data type
+                type_front = "ARRAY<"
+                type_back = ">"
+                for i in range(1, col.array_dimensional):
+                    type_front += "STRUCT<dimension_{} ARRAY<".format(i)
+                    type_back += ">>"
+
+                type = "{}{}{}".format(type_front, col.bigquery_standard_data_type, type_back)
+                not_null = ""
+
+            # cols_defs.append("{name} {type}{not_null}".format(
+            cols_defs.append("{name} {type}{not_null}{description}".format(
+                name=col_name,
+                type=type,
+                not_null=not_null,
+                description=' OPTIONS (description = "{}")'.format(col.description.replace('"', '\\"')) if col.description is not None else "",
+            ))
+
+        return textwrap.dedent(
+            """\
+            #standardSQL
+            CREATE TABLE `project.{dataset}.{table}`
+            (
+              {colmns_define}
+            )""").format(
+            dataset=dataset,
+            table=self.get_name(name_case),
+            colmns_define=",\n  ".join(cols_defs),
+        )
+
 
 class DdlParse(DdlParseBase):
     """DDL parser"""
 
     _LPAR, _RPAR, _COMMA, _SEMICOLON, _DOT, _DOUBLEQUOTE, _BACKQUOTE, _SPACE = map(Suppress, "(),;.\"` ")
-    _CREATE, _TABLE, _TEMP, _CONSTRAINT, _NOT_NULL, _PRIMARY_KEY, _UNIQUE, _UNIQUE_KEY, _KEY, _CHAR_SEMANTICS, _BYTE_SEMANTICS = \
-        map(CaselessKeyword, "CREATE, TABLE, TEMP, CONSTRAINT, NOT NULL, PRIMARY KEY, UNIQUE, UNIQUE KEY, KEY, CHAR, BYTE".replace(", ", ",").split(","))
+    _CREATE, _TABLE, _TEMP, _CONSTRAINT, _NOT_NULL, _PRIMARY_KEY, _UNIQUE, _UNIQUE_KEY, _FOREIGN_KEY, _REFERENCES, _KEY, _CHAR_SEMANTICS, _BYTE_SEMANTICS = \
+        map(CaselessKeyword, "CREATE, TABLE, TEMP, CONSTRAINT, NOT NULL, PRIMARY KEY, UNIQUE, UNIQUE KEY, FOREIGN KEY, REFERENCES, KEY, CHAR, BYTE".replace(", ", ",").split(","))
     _SUPPRESS_QUOTE = _BACKQUOTE | _DOUBLEQUOTE
 
     _COMMENT = Suppress("--" + Regex(r".+"))
 
+    _COLUMN_CONSTRAINT_BASE = r"""
+    (?!--)
+    (
+        (
+            \s*\b(?:NOT\s+)NULL?\b
+        )?
+        (
+            \s*\bAUTO_INCREMENT\b
+        )?(
+            \s*\b(UNIQUE|PRIMARY)(?:\s+KEY)?\b
+        )?
+        (
+            \s*\bDEFAULT\b\s+(
+                ([A-Za-z0-9_\.\'\" -]|[^\x01-\x7E])*\:\:[A-Za-z0-9\[\]]+
+                |
+                \'(\\\'|[^\']|,)+\'
+                |
+                \"(\\\"|[^\"]|,)+\"
+                |
+                [^,]+
+            )
+        )?
+        (
+            \s*\bCOMMENT\b\s+(
+                \'(\\\'|[^\']|,)+\'
+                |
+                \"(\\\"|[^\"]|,)+\"
+                |
+                [^,]+
+            )
+        )?
+    )
+    """
+
+    _COLUMN_CONSTRAINT = re.sub(r"(^\s+|\n)", r"", _COLUMN_CONSTRAINT_BASE, flags=re.MULTILINE)
+
+
     _CREATE_TABLE_STATEMENT = Suppress(_CREATE) + Optional(_TEMP)("temp") + Suppress(_TABLE) + Optional(Suppress(CaselessKeyword("IF NOT EXISTS"))) \
         + Optional(_SUPPRESS_QUOTE) + Optional(Word(alphanums+"_")("schema") + Optional(_SUPPRESS_QUOTE) + _DOT + Optional(_SUPPRESS_QUOTE)) + Word(alphanums+"_<>")("table") + Optional(_SUPPRESS_QUOTE) \
         + _LPAR \
         + delimitedList(
             OneOrMore(
+                _COMMENT
+                |
                 # Ignore Index
                 Suppress(_KEY + Word(alphanums+"_'`() "))
                 |
                 Group(
                     Optional(Suppress(_CONSTRAINT) + Optional(_SUPPRESS_QUOTE) + Word(alphanums+"_")("name") + Optional(_SUPPRESS_QUOTE))
-                    + (_PRIMARY_KEY ^ _UNIQUE ^ _UNIQUE_KEY ^ _NOT_NULL)("type")
-                    + Optional(_SUPPRESS_QUOTE) + Optional(Word(alphanums+"_"))("name") + Optional(_SUPPRESS_QUOTE)
-                    + _LPAR + Group(delimitedList(Optional(_SUPPRESS_QUOTE) + Word(alphanums+"_") + Optional(_SUPPRESS_QUOTE)))("constraint_columns") + _RPAR
+                    + (
+                        (
+                            (_PRIMARY_KEY ^ _UNIQUE ^ _UNIQUE_KEY ^ _NOT_NULL)("type")
+                            + Optional(_SUPPRESS_QUOTE) + Optional(Word(alphanums+"_"))("name") + Optional(_SUPPRESS_QUOTE)
+                            + _LPAR + Group(delimitedList(Optional(_SUPPRESS_QUOTE) + Word(alphanums+"_") + Optional(_SUPPRESS_QUOTE)))("constraint_columns") + _RPAR
+                        )
+                        |
+                        (
+                            (_FOREIGN_KEY)("type")
+                            + _LPAR + Group(delimitedList(Optional(_SUPPRESS_QUOTE) + Word(alphanums+"_") + Optional(_SUPPRESS_QUOTE)))("constraint_columns") + _RPAR
+                            + Optional(Suppress(_REFERENCES)
+                                + Optional(_SUPPRESS_QUOTE) + Word(alphanums+"_")("references_table") + Optional(_SUPPRESS_QUOTE)
+                                + _LPAR + Group(delimitedList(Optional(_SUPPRESS_QUOTE) + Word(alphanums+"_") + Optional(_SUPPRESS_QUOTE)))("references_columns") + _RPAR
+                            )
+                        )
+                    )
                 )("constraint")
                 |
                 Group(
-                    Optional(_COMMENT)
-                    + Optional(_SUPPRESS_QUOTE) + Word(alphanums+"_")("name") + Optional(_SUPPRESS_QUOTE)
+                    Optional(_SUPPRESS_QUOTE) + Word(alphanums+"_")("name") + Optional(_SUPPRESS_QUOTE)
                     + Group(
                           Word(alphanums+"_")
-                        + Optional(CaselessKeyword("WITHOUT TIME ZONE") ^ CaselessKeyword("WITH TIME ZONE") ^ CaselessKeyword("PRECISION"))
-                        + Optional(_LPAR + Regex(r"\d+\s*,*\s*\d*") + Optional(Suppress(_CHAR_SEMANTICS | _BYTE_SEMANTICS)) + _RPAR)
+                        + Optional(CaselessKeyword("WITHOUT TIME ZONE") ^ CaselessKeyword("WITH TIME ZONE") ^ CaselessKeyword("PRECISION") ^ CaselessKeyword("VARYING"))
+                        + Optional(_LPAR + Regex(r"[\d\*]+\s*,*\s*\d*") + Optional(Suppress(_CHAR_SEMANTICS | _BYTE_SEMANTICS)) + _RPAR)
                         )("type")
-                    + Optional(Word(alphanums+"_': -"))("constraint")
+                    + Optional(Word(r"\[\]"))("array_brackets")
+                    + Optional(Regex(_COLUMN_CONSTRAINT, re.IGNORECASE))("constraint")
                 )("column")
+                |
+                _COMMENT
             )
         )("columns")
 
@@ -356,6 +588,7 @@ class DdlParse(DdlParseBase):
     def source_database(self):
         """
         Source database option
+
         :param source_database: enum DdlParse.DATABASE
         """
         return super().source_database
@@ -406,10 +639,9 @@ class DdlParse(DdlParseBase):
                 # add column
                 col = self._table.columns.append(
                     column_name=ret_col["name"],
-                    data_type_array=ret_col["type"])
-
-                if "constraint" in ret_col:
-                    col.constraint = ret_col["constraint"]
+                    data_type_array=ret_col["type"],
+                    array_brackets=ret_col['array_brackets'] if "array_brackets" in ret_col else None,
+                    constraint=ret_col['constraint'] if "constraint" in ret_col else None)
 
             elif ret_col.getName() == "constraint":
                 # set column constraint
